from ..default import *
from .. import structure
from .basic import * 
from ..statistics import stats_methods

import scipy.fftpack
from scipy import signal

comparison_params = dict(test=stats_methods.t_test, win='1ms', method='mean', sig_limit=0.05, need_fdr=False)

def Spectrum(self, compare=False, freq_span=(0,30), target='power', comparison_params=comparison_params):
    # with the decorator, we can just focuse on case data instead of batch/collection data
    @self.iter('average')
    def to_spetrum(case_raw_data):
        def fft(name,data):
            N = data.shape[1]

            fft_result = scipy.fftpack.fft(data,axis=1)
            if target=='power':
                fft_result = 2.0/N * np.abs(fft_result)
            elif target=='phase':
                fft_result = np.angle(fft_result)
            else:
                raise Exception('Please set the parameter "target" as "power", or "phase".')
            fft_result = fft_result[:,:N//2]
            if freq_span[0]==0:
                fft_result[:,0] = 0

            index = pd.MultiIndex.from_tuples([name], names=['subject','condition_group','channel_group'])

            fft_result = pd.DataFrame(fft_result, index=index,
                columns=np.linspace(0, self.info['sample_rate'], N//2)) # resolution: sr/N
            fft_result = fft_result.loc[:, freq_span[0]:freq_span[1]] 
            
            fft_result.columns.name = 'frequency'
            return fft_result

        Spectrum_df = convert(case_raw_data, ['subject', 'condition_group', 'channel_group'], fft)
        return Spectrum_df
    
    spetrum_batch = to_spetrum()
    if compare:
        stats_data = stats_compare(spetrum_batch, comparison_params, levels='frequency', between='condition_group', in_group='subject')
    else:
        stats_data = None

    default_plot_params = dict(plot_type=['direct', 'spectrum'], y_title='Spectrum', err_style='ci_band',
                               color="Set1", style='darkgrid', compare=compare, win=comparison_params['win'], sig_limit=0.05)
    return structure.Analyzed_data('Spectrum', spetrum_batch, stats_data, default_plot_params)

# grand average
def Time_frequency(self, compare=False, freq_span=(0, 30)):
    freqs = np.arange(freq_span[0], freq_span[1])
    if freq_span[0] == 0:
        widths = freqs+0.001
    else:
        widths = freqs

    # with the decorator, we can just focuse on case data instead of batch/collection data
    @self.iter('average')
    def to_tf(case_raw_data):
        def cwt(name, data):
            cwt_result = signal.cwt(
                np.array(data)[0], signal.ricker, widths=widths)
            cwt_result = pd.DataFrame(cwt_result, columns=data.columns)
            cwt_result.index = pd.MultiIndex.from_tuples(
                [(name[0], name[1], i) for i in freqs[::-1]], names=('condition_group', 'channel_group', 'freq'))
            return cwt_result

        tf_df = convert(case_raw_data, ['condition_group', 'channel_group'], cwt)

        for level in ['condition_group', 'channel_group']:
            if len(tf_df.index.get_level_values(level).unique()) == 2:
                tf_df = [i for ind,i in tf_df.groupby(level=level)]
                tf_df[0].index = tf_df[0].index.droplevel(level)
                tf_df[1].index = tf_df[1].index.droplevel(level)
                tf_df = tf_df[0] - tf_df[1]
        
        tf_df.index = tf_df.index.get_level_values('freq')

        return tf_df

    @self.iter('average')
    def to_tf_group_level(case_raw_data):
        def cwt(name, data):
            cwt_result = signal.cwt(np.array(data)[0], signal.ricker, widths=widths)
            cwt_result = pd.DataFrame(cwt_result, columns=data.columns)
            cwt_result.index = pd.MultiIndex.from_tuples(
                [(name[0], name[1], name[2], i) for i in freqs[::-1]], names=('condition_group', 'channel_group', 'subject',  'freq'))
            return cwt_result

        tf_df = convert(case_raw_data, ['condition_group', 'channel_group', 'subject'], cwt)

        return tf_df
    
    tf_batch = to_tf()

    if compare:
        tf_group_level_batch = to_tf_group_level()
        stats_data = stats_compare(tf_group_level_batch, comparison_params, levels=['time', 'freq'], between='condition_group', in_group='subject')
    else:
        stats_data = None

    default_plot_params = dict(plot_type=['direct', 'heatmap'], x_title='time', y_title='frequency', compare=compare, sig_limit=0.05,
        color="RdBu_r", style='white', grid=False, cbar_title='Power')
    return structure.Analyzed_data('Time Frequency', tf_batch, stats_data, default_plot_params)



from ..default import *
from .. import structure
from .basic import * 
from ..statistics import stats_methods
from scipy import signal

def topography(self, step_size='1ms', win_size='1ms', sample='mean', sig_limit=0.05):
    @self.iter('average')
    def to_topo(case_raw_data):
        check_availability(case_raw_data, 'channel', '>1')

        data_with_subject = case_raw_data.mean(level=['subject', 'condition_group', 'channel'])
        data_with_subject = sampling(data_with_subject, step_size, win_size, sample)

        data_with_subject.columns = data_with_subject.columns.get_level_values(1)  # remvoe "time_group"
        
        if len(data_with_subject.index.get_level_values('condition_group').unique()) == 2:
            topo_result = subtract(data_with_subject, between='condition_group', align='channel')

            stats_result = roll_on_levels_and_compare(data_with_subject, stats_methods.t_test, levels=['time', 'channel'], between='condition_group', in_group='subject', prograssbar=False)
            return topo_result, stats_result

        else:
            topo_result = data_with_subject.mean(level=['condition_group', 'channel'])
            return topo_result, None
    
    topo_batch, stats_data = to_topo()

    minmax = (topo_batch.min().min(), topo_batch.max().max())

    default_plot_params = dict(title='Topography', plot_type=['matrix', 'topograph'], zlim=minmax, color=plt.cm.jet, cbar_title='uV', chan_locs=self.info['xy_locs'], sig_limit=sig_limit, x_title='time', y_title='condition_group')

    return structure.Analyzed_data('Topography', topo_batch, stats_data, default_plot_params=default_plot_params)

def frequency_topography(self, step_size='1ms', win_size='1ms', sample='mean', sig_limit=0.05, target=10):
    if isinstance(target, (int, float)):
        freqs = [target]
    elif isinstance(target, list) and len(target) == 2 \
            and isinstance(target[0], (int, float)) and isinstance(target[1], (int, float)):
        freqs = np.arange(target[0], target[1])
    else:
        raise Exception(
            'Unsupported value for "target". The value should be a number, or a list of two numbers')
    
    if freqs[0] == 0: freqs += 0.001
    
    def cwt(name, data):
        cwt_result = signal.cwt(
            np.array(data)[0], signal.ricker, widths=freqs).mean(axis=0)
        cwt_result = pd.DataFrame(
            [cwt_result], index=data.index, columns=data.columns)
        return cwt_result

    @self.iter('average')
    def to_topo(case_raw_data):
        data_without_subject = convert(case_raw_data, ['condition_group', 'channel'], cwt)
        data_without_subject = sampling(data_without_subject, step_size, win_size, sample)

        data_without_subject.columns = data_without_subject.columns.get_level_values(1)  # remvoe "time_group"

        if len(data_without_subject.index.get_level_values('condition_group').unique()) == 2:
            data_with_subject = convert(case_raw_data, ['subject', 'condition_group', 'channel'], cwt)
            data_with_subject = sampling(data_with_subject, step_size, win_size, sample)

            data_with_subject.columns = data_with_subject.columns.get_level_values(1)  # remvoe "time_group"  

            topo_result = subtract(
                data_with_subject, between='condition_group', align='channel')

            stats_result = roll_on_levels_and_compare(data_with_subject, stats_methods.t_test, levels=['time', 'channel'], between='condition_group', in_group='subject', prograssbar=False)

            return topo_result, stats_result
        else:
            topo_result = data_without_subject
            return topo_result, None

    topo_batch, stats_data = to_topo()
    minmax = (topo_batch.min().min(), topo_batch.max().max())

    default_plot_params = dict(title='Topography', plot_type=['matrix', 'topograph'], zlim=minmax, color=plt.cm.jet, cbar_title='Power', chan_locs=self.info['xy_locs'], sig_limit=sig_limit, x_title='time', y_title='condition_group')

    return structure.Analyzed_data('Topography', topo_batch, stats_data, default_plot_params=default_plot_params)

def significant_channels_count(self, step_size='1ms', win_size='1ms', sample='mean', sig_limit=0.05):
    @self.iter('average')
    def to_signif(case_raw_data):
        case_raw_data = sampling(case_raw_data, step_size, win_size, sample)
        data_with_subject = case_raw_data.mean(level=['subject','condition_group','channel'])
        check_availability(data_with_subject, 'condition_group','==2')

        stats_result = roll_on_levels_and_compare(data_with_subject, stats_methods.t_test, levels=['time','channel'], between='condition_group', in_group='subject',prograssbar=True)
  
        stats_result = stats_result.stack('time').unstack('channel').apply(lambda x:sum(x<sig_limit),axis=1).unstack('time')

        return stats_result

    signif_batch = to_signif()

    default_plot_params = dict(title='significant_channels_count',plot_type=['direct','heatmap'], x_len=12, color=sns.cubehelix_palette(light=1, as_cmap=True), x_title='time', y_title='condition_group',cbar_title='Count')

    return structure.Analyzed_data('significant_channels_count', signif_batch, default_plot_params=default_plot_params)
    
